from plain import forms
from plain.auth import get_user_model
from plain.exceptions import ValidationError
from plain.models.forms import ModelForm

from .core import check_user_password
from .hashers import check_password

# class PasswordResetForm(forms.Form):
#     email = forms.EmailField(
#         # label="Email",
#         max_length=254,
#         # widget=forms.EmailInput(attrs={"autocomplete": "email"}),
#     )

#     def send_mail(
#         self,
#         subject_template_name,
#         email_template_name,
#         context,
#         from_email,
#         to_email,
#         html_email_template_name=None,
#     ):
#         from plain.mail import EmailMultiAlternatives

#         """
#         Send a plain.mail.EmailMultiAlternatives to `to_email`.
#         """
#         template = Template(subject_template_name)
#         subject = template.render(context)
#         # Email subject *must not* contain newlines
#         subject = "".join(subject.splitlines())
#         template = Template(email_template_name)
#         body = template.render(context)

#         email_message = EmailMultiAlternatives(subject, body, from_email, [to_email])
#         if html_email_template_name is not None:
#             template = Template(html_email_template_name)
#             html_email = template.render(context)
#             email_message.attach_alternative(html_email, "text/html")

#         email_message.send()

#     def get_users(self, email):
#         """Given an email, return matching user(s) who should receive a reset.

#         This allows subclasses to more easily customize the default policies
#         that prevent inactive users and users with unusable passwords from
#         resetting their password.
#         """
#         active_users = get_user_model()._default_manager.filter(email__iexact=email)
#         return (u for u in active_users if _unicode_ci_compare(email, u.email))

#     def save(
#         self,
#         subject_template_name="auth/password_reset_subject.txt",
#         email_template_name="auth/password_reset_email.html",
#         use_https=False,
#         token_generator=default_token_generator,
#         from_email=None,
#         html_email_template_name=None,
#         extra_email_context=None,
#     ):
#         """
#         Generate a one-use only link for resetting password and send it to the
#         user.
#         """
#         email = self.cleaned_data["email"]
#         for user in self.get_users(email):
#             context = {
#                 "email": user.email,
#                 "uid": urlsafe_base64_encode(force_bytes(user.pk)),
#                 "user": user,
#                 "token": token_generator.make_token(user),
#                 "protocol": "https" if use_https else "http",
#                 **(extra_email_context or {}),
#             }
#             self.send_mail(
#                 subject_template_name,
#                 email_template_name,
#                 context,
#                 from_email,
#                 user.email,
#                 html_email_template_name=html_email_template_name,
#             )


# class SetPasswordForm(forms.Form):
#     """
#     A form that lets a user set their password without entering the old
#     password
#     """

#     error_messages = {
#         "password_mismatch": "The two password fields didn’t match.",
#     }
#     new_password1 = forms.CharField(
#         # label="New password",
#         # widget=forms.PasswordInput(attrs={"autocomplete": "new-password"}),
#         strip=False,
#         # help_text=validators.password_validators_help_text_html(),
#     )
#     new_password2 = forms.CharField(
#         # label="New password confirmation",
#         strip=False,
#         # widget=forms.PasswordInput(attrs={"autocomplete": "new-password"}),
#     )

#     def __init__(self, user, *args, **kwargs):
#         self.user = user
#         super().__init__(*args, **kwargs)

#     def clean_new_password2(self):
#         password1 = self.cleaned_data.get("new_password1")
#         password2 = self.cleaned_data.get("new_password2")
#         if password1 and password2 and password1 != password2:
#             raise ValidationError(
#                 self.error_messages["password_mismatch"],
#                 code="password_mismatch",
#             )
#         validators.validate_password(password2, self.user)
#         return password2

#     def save(self, commit=True):
#         password = self.cleaned_data["new_password1"]
#         self.user.set_password(password)
#         if commit:
#             self.user.save()
#         return self.user


# class PasswordChangeForm(SetPasswordForm):
#     """
#     A form that lets a user change their password by entering their old
#     password.
#     """

#     error_messages = {
#         **SetPasswordForm.error_messages,
#         "password_incorrect": "Your old password was entered incorrectly. Please enter it again.",
#     }
#     old_password = forms.CharField(
#         # label="Old password",
#         strip=False,
#         # widget=forms.PasswordInput(
#         #     attrs={"autocomplete": "current-password", "autofocus": True}
#         # ),
#     )

#     field_order = ["old_password", "new_password1", "new_password2"]

#     def clean_old_password(self):
#         """
#         Validate that the old_password field is correct.
#         """
#         old_password = self.cleaned_data["old_password"]
#         if not self.user.check_password(old_password):
#             raise ValidationError(
#                 self.error_messages["password_incorrect"],
#                 code="password_incorrect",
#             )
#         return old_password


class PasswordLoginForm(forms.Form):
    email = forms.EmailField(max_length=150)
    password = forms.CharField(strip=False)

    def clean(self):
        User = get_user_model()

        email = self.cleaned_data.get("email")
        password = self.cleaned_data.get("password")

        if email and password:
            try:
                # The vast majority of users won't have a case-sensitive email, so we act that way
                user = User.objects.get(email__iexact=email)
            except User.DoesNotExist:
                # Run the default password hasher once to reduce the timing
                # difference between an existing and a nonexistent user (django #20760).
                check_password(password, "")

                raise ValidationError(
                    "Please enter a correct email and password. Note that both fields may be case-sensitive.",
                    code="invalid_login",
                )

            if not check_user_password(user, password):
                raise ValidationError(
                    "Please enter a correct email and password. Note that both fields may be case-sensitive.",
                    code="invalid_login",
                )

            self._user = user

        return self.cleaned_data

    def get_user(self):
        return self._user


class PasswordSignupForm(ModelForm):
    confirm_password = forms.CharField(strip=False)

    class Meta:
        model = get_user_model()
        fields = ("email", "password")

    def clean(self):
        cleaned_data = super().clean()
        password = cleaned_data.get("password")
        confirm_password = cleaned_data.get("confirm_password")
        if password and confirm_password and password != confirm_password:
            raise ValidationError(
                "The two password fields didn't match.",
                code="password_mismatch",
            )
        return cleaned_data



from plain.urls import Router, path

from . import views


class PasswordsRouter(Router):
    namespace = "passwords"
    urls = [
        path("password_change/", views.PasswordChangeView, name="password_change"),
        path(
            "password_change/done/",
            views.PasswordChangeDoneView,
            name="password_change_done",
        ),
        path("password_reset/", views.PasswordResetView, name="password_reset"),
        path(
            "password_reset/done/",
            views.PasswordResetDoneView,
            name="password_reset_done",
        ),
        path(
            "reset/<uidb64>/<token>/",
            views.PasswordResetConfirmView,
            name="password_reset_confirm",
        ),
        path(
            "reset/done/",
            views.PasswordResetCompleteView,
            name="password_reset_complete",
        ),
    ]



# Avoid shadowing the login() and logout() views below.
from plain.auth.sessions import USER_HASH_SESSION_KEY, get_session_auth_hash
from plain.auth.sessions import login as auth_login
from plain.http import (
    ResponseRedirect,
)
from plain.views import CreateView, FormView

from .forms import (
    # PasswordChangeForm,
    # PasswordResetForm,
    # SetPasswordForm,
    PasswordLoginForm,
    PasswordSignupForm,
)


def update_session_auth_hash(request, user):
    """
    Updating a user's password logs out all sessions for the user.

    Take the current request and the updated user object from which the new
    session hash will be derived and update the session hash appropriately to
    prevent a password change from logging out the session from which the
    password was changed.
    """
    request.session.cycle_key()
    if request.user == user:
        request.session[USER_HASH_SESSION_KEY] = get_session_auth_hash(user)


# Class-based password reset views
# - PasswordResetView sends the mail
# - PasswordResetDoneView shows a success message for the above
# - PasswordResetConfirmView checks the link the user clicked and
#   prompts for a new password
# - PasswordResetCompleteView shows a success message for the above


# class PasswordContextMixin:
#     extra_context = None

#     def get_template_context(self):
#         context = super().get_template_context()
#         context.update(
#             {"title": self.title, "subtitle": None, **(self.extra_context or {})}
#         )
#         return context


# class PasswordResetView(PasswordContextMixin, FormView):
#     email_template_name = "auth/password_reset_email.html"
#     extra_email_context = None
#     form_class = PasswordResetForm
#     from_email = None
#     html_email_template_name = None
#     subject_template_name = "auth/password_reset_subject.txt"
#     success_url = reverse_lazy("password_reset_done")
#     template_name = "auth/password_reset_form.html"
#     title = "Password reset"
#     token_generator = default_token_generator

#     def form_valid(self, form):
#         opts = {
#             "use_https": self.request.is_https(),
#             "token_generator": self.token_generator,
#             "from_email": self.from_email,
#             "email_template_name": self.email_template_name,
#             "subject_template_name": self.subject_template_name,
#             "html_email_template_name": self.html_email_template_name,
#             "extra_email_context": self.extra_email_context,
#         }
#         form.save(**opts)
#         return super().form_valid(form)


# INTERNAL_RESET_SESSION_TOKEN = "_password_reset_token"


# class PasswordResetDoneView(PasswordContextMixin, TemplateView):
#     template_name = "auth/password_reset_done.html"
#     title = "Password reset sent"


# class PasswordResetConfirmView(PasswordContextMixin, FormView):
#     form_class = SetPasswordForm
#     post_reset_login = False
#     post_reset_login_backend = None
#     reset_url_token = "set-password"
#     success_url = reverse_lazy("password_reset_complete")
#     template_name = "auth/password_reset_confirm.html"
#     title = "Enter new password"
#     token_generator = default_token_generator

#     def get_response(self):
#         if "uidb64" not in self.url_kwargs or "token" not in self.url_kwargs:
#             raise ImproperlyConfigured(
#                 "The URL path must contain 'uidb64' and 'token' parameters."
#             )

#         self.validlink = False
#         self.user = self.get_user(self.url_kwargs["uidb64"])

#         if self.user is not None:
#             token = self.url_kwargs["token"]
#             if token == self.reset_url_token:
#                 session_token = self.request.session.get(INTERNAL_RESET_SESSION_TOKEN)
#                 if self.token_generator.check_token(self.user, session_token):
#                     # If the token is valid, display the password reset form.
#                     self.validlink = True
#                     response = super().get_response()
#                     add_never_cache_headers(response)
#                     return response
#             else:
#                 if self.token_generator.check_token(self.user, token):
#                     # Store the token in the session and redirect to the
#                     # password reset form at a URL without the token. That
#                     # avoids the possibility of leaking the token in the
#                     # HTTP Referer header.
#                     self.request.session[INTERNAL_RESET_SESSION_TOKEN] = token
#                     redirect_url = self.request.path.replace(
#                         token, self.reset_url_token
#                     )
#                     response = ResponseRedirect(redirect_url)
#                     add_never_cache_headers(response)
#                     return response

#         # Display the "Password reset unsuccessful" page.
#         response = self.render_to_response(self.get_template_context())
#         add_never_cache_headers(response)
#         return response

#     def get_user(self, uidb64):
#         UserModel = get_user_model()
#         try:
#             # urlsafe_base64_decode() decodes to bytestring
#             uid = urlsafe_base64_decode(uidb64).decode()
#             user = UserModel._default_manager.get(pk=uid)
#         except (
#             TypeError,
#             ValueError,
#             OverflowError,
#             UserModel.DoesNotExist,
#             ValidationError,
#         ):
#             user = None
#         return user

#     def get_form_kwargs(self):
#         kwargs = super().get_form_kwargs()
#         kwargs["user"] = self.user
#         return kwargs

#     def form_valid(self, form):
#         user = form.save()
#         del self.request.session[INTERNAL_RESET_SESSION_TOKEN]
#         if self.post_reset_login:
#             auth_login(self.request, user, self.post_reset_login_backend)
#         return super().form_valid(form)

#     def get_template_context(self):
#         context = super().get_template_context()
#         if self.validlink:
#             context["validlink"] = True
#         else:
#             context.update(
#                 {
#                     "form": None,
#                     "title": "Password reset unsuccessful",
#                     "validlink": False,
#                 }
#             )
#         return context


# class PasswordResetCompleteView(PasswordContextMixin, TemplateView):
#     template_name = "auth/password_reset_complete.html"
#     title = "Password reset complete"

#     def get_template_context(self):
#         context = super().get_template_context()
#         context["login_url"] = resolve_url(settings.AUTH_LOGIN_URL)
#         return context


# class PasswordChangeView(PasswordContextMixin, AuthViewMixin, FormView):
#     form_class = PasswordChangeForm
#     success_url = reverse_lazy("password_change_done")
#     template_name = "auth/password_change_form.html"
#     title = "Password change"
#     login_required = True

#     def get_form_kwargs(self):
#         kwargs = super().get_form_kwargs()
#         kwargs["user"] = self.request.user
#         return kwargs

#     def form_valid(self, form):
#         form.save()
#         # Updating the password logs out all other sessions for the user
#         # except the current one.
#         update_session_auth_hash(self.request, form.user)
#         return super().form_valid(form)


# class PasswordChangeDoneView(PasswordContextMixin, AuthViewMixin, TemplateView):
#     template_name = "auth/password_change_done.html"
#     title = "Password change successful"
#     login_required = True


class PasswordLoginView(FormView):
    form_class = PasswordLoginForm
    success_url = "/"

    def get(self):
        # Redirect if the user is already logged in
        if self.request.user:
            return ResponseRedirect(self.success_url)

        return super().get()

    def form_valid(self, form):
        # Log the user in and redirect
        auth_login(self.request, form.get_user())

        return super().form_valid(form)


class PasswordSignupView(CreateView):
    form_class = PasswordSignupForm
    success_url = "/"

    def form_valid(self, form):
        # # Log the user in and redirect
        # auth_login(self.request, form.save())

        return super().form_valid(form)



# The number of seconds a password reset link is valid for (default: 3 days).
PASSWORD_RESET_TIMEOUT: int = 60 * 60 * 24 * 3

# the first hasher in this list is the preferred algorithm.  any
# password using different algorithms will be converted automatically
# upon login
PASSWORD_HASHERS: list = [
    "plain.passwords.hashers.PBKDF2PasswordHasher",
    "plain.passwords.hashers.PBKDF2SHA1PasswordHasher",
    "plain.passwords.hashers.Argon2PasswordHasher",
    "plain.passwords.hashers.BCryptSHA256PasswordHasher",
    "plain.passwords.hashers.ScryptPasswordHasher",
]



import gzip
from pathlib import Path

from plain.exceptions import (
    ValidationError,
)
from plain.utils.deconstruct import deconstructible
from plain.utils.functional import cached_property
from plain.utils.text import pluralize


@deconstructible
class MinimumLengthValidator:
    """
    Validate that the password is of a minimum length.
    """

    def __init__(self, min_length=8):
        self.min_length = min_length

    def __call__(self, password):
        if len(password) < self.min_length:
            raise ValidationError(
                pluralize(
                    "This password is too short. It must contain at least "
                    "%(min_length)d character.",
                    "This password is too short. It must contain at least "
                    "%(min_length)d characters.",
                    self.min_length,
                ),
                code="password_too_short",
                params={"min_length": self.min_length},
            )


# def exceeds_maximum_length_ratio(password, max_similarity, value):
#     """
#     Test that value is within a reasonable range of password.

#     The following ratio calculations are based on testing SequenceMatcher like
#     this:

#     for i in range(0,6):
#       print(10**i, SequenceMatcher(a='A', b='A'*(10**i)).quick_ratio())

#     which yields:

#     1 1.0
#     10 0.18181818181818182
#     100 0.019801980198019802
#     1000 0.001998001998001998
#     10000 0.00019998000199980003
#     100000 1.999980000199998e-05

#     This means a length_ratio of 10 should never yield a similarity higher than
#     0.2, for 100 this is down to 0.02 and for 1000 it is 0.002. This can be
#     calculated via 2 / length_ratio. As a result we avoid the potentially
#     expensive sequence matching.
#     """
#     pwd_len = len(password)
#     length_bound_similarity = max_similarity / 2 * pwd_len
#     value_len = len(value)
#     return pwd_len >= 10 * value_len and value_len < length_bound_similarity


# @deconstructible
# class UserAttributeSimilarityValidator:
#     """
#     Validate that the password is sufficiently different from the user's
#     attributes.

#     If no specific attributes are provided, look at a sensible list of
#     defaults. Attributes that don't exist are ignored. Comparison is made to
#     not only the full attribute value, but also its components, so that, for
#     example, a password is validated against either part of an email address,
#     as well as the full address.
#     """

#     DEFAULT_USER_ATTRIBUTES = ("username", "email")

#     def __init__(self, user_attributes=DEFAULT_USER_ATTRIBUTES, max_similarity=0.7):
#         self.user_attributes = user_attributes
#         if max_similarity < 0.1:
#             raise ValueError("max_similarity must be at least 0.1")
#         self.max_similarity = max_similarity

#     def validate(self, password, user=None):
#         if not user:
#             return

#         password = password.lower()
#         for attribute_name in self.user_attributes:
#             value = getattr(user, attribute_name, None)
#             if not value or not isinstance(value, str):
#                 continue
#             value_lower = value.lower()
#             value_parts = re.split(r"\W+", value_lower) + [value_lower]
#             for value_part in value_parts:
#                 if exceeds_maximum_length_ratio(
#                     password, self.max_similarity, value_part
#                 ):
#                     continue
#                 if (
#                     SequenceMatcher(a=password, b=value_part).quick_ratio()
#                     >= self.max_similarity
#                 ):
#                     try:
#                         verbose_name = str(
#                             user._meta.get_field(attribute_name).verbose_name
#                         )
#                     except FieldDoesNotExist:
#                         verbose_name = attribute_name
#                     raise ValidationError(
#                         "The password is too similar to the %(verbose_name)s.",
#                         code="password_too_similar",
#                         params={"verbose_name": verbose_name},
#                     )

#     def get_help_text(self):
#         return "Your password can’t be too similar to your other personal information."


@deconstructible
class CommonPasswordValidator:
    """
    Validate that the password is not a common password.

    The password is rejected if it occurs in a provided list of passwords,
    which may be gzipped. The list Plain ships with contains 20000 common
    passwords (lowercased and deduplicated), created by Royce Williams:
    https://gist.github.com/roycewilliams/226886fd01572964e1431ac8afc999ce
    The password list must be lowercased to match the comparison in validate().
    """

    @cached_property
    def DEFAULT_PASSWORD_LIST_PATH(self):
        return Path(__file__).resolve().parent / "common-passwords.txt.gz"

    def __init__(self, password_list_path=DEFAULT_PASSWORD_LIST_PATH):
        if password_list_path is CommonPasswordValidator.DEFAULT_PASSWORD_LIST_PATH:
            password_list_path = self.DEFAULT_PASSWORD_LIST_PATH
        try:
            with gzip.open(password_list_path, "rt", encoding="utf-8") as f:
                self.passwords = {x.strip() for x in f}
        except OSError:
            with open(password_list_path) as f:
                self.passwords = {x.strip() for x in f}

    def __call__(self, password):
        if password.lower().strip() in self.passwords:
            raise ValidationError(
                "This password is too common.",
                code="password_too_common",
            )


@deconstructible
class NumericPasswordValidator:
    """
    Validate that the password is not entirely numeric.
    """

    def __call__(self, password):
        if password.isdigit():
            raise ValidationError(
                "This password is entirely numeric.",
                code="password_entirely_numeric",
            )
